//    Austin is a program for updating projects kept in XML files.
//    Copyright (C) 2000  Joseph Gregorio
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#pragma warning(disable:4786)

#include <algorithm>
#include <cassert>  
#include <sstream>
#include <ctime>

#include "JulianDay.h"

using namespace std;

//----------------------------------------------------------------------------------------
// Helper function. Calculates the julian day number based on the given date
//
double caluculateJulianDayNumber(const struct tm * date)
{
	double returnValue = 0;
	int wMonth 	= date->tm_mon+1;
	int wDay 	= date->tm_mday;
	int wYear 	= date->tm_year+1900;

	// Check for invalid dates
	if((wMonth < 1) || (wMonth > 12) || (wDay < 1) || (wDay > 31) || (wYear < 400)) {
		assert(true);
		return DEFUALT_JULIAN_DAY_NUMBER;
	}

	if(wMonth < 3) {
		wMonth += 12;
		wYear --;
	}
	int A, B, C, E, F;
	A = wYear/100;
	B = A/4;
	C = 2 - A + B;
	E = 365.25 * (wYear + 4716);
	F = 30.6001 * (wMonth + 1);
	returnValue = C + wDay + E + F - 1524.5;

	return returnValue;
}

//----------------------------------------------------------------------------------------
// Helper function.  Returns the date based on the given Julian Day Number value.
//
void caluculateDate(double julianDayNumber, struct tm * date) 
{
	int Z, W, X, A, B, C, D, E, F;

	int wMonth = 0;
	int wYear = 0;
	int wDay = 0;

	Z = julianDayNumber + 0.5;
	W = (Z - 1867216.25)/36524.25;
	X = W/4;
	A = Z + 1 + W - X;
	B = A + 1524;
	C = (B - 122.1)/365.25;
	D = 365.25 * C;
	E = (B - D) / 30.6001;
	F = 30.6001 * E;
	wDay = B - D - F;
	if(E <= 13) {
		wMonth = E - 1;
		wYear = C - 4716;
	}
	else {  // (Jan. or Feb.)
		wMonth = E - 13;
		wYear = C - 4715;
	}

	memset(date, 0, sizeof(struct tm));
	date->tm_mon = wMonth-1;
	date->tm_mday = wDay;
	date->tm_year = wYear-1900;
}

JulianDay::JulianDay() 
{
	time_t t;
	time(&t);
	julianDayNumber_ = caluculateJulianDayNumber(localtime(&t));
}

JulianDay::JulianDay(const struct tm * date)
{
	julianDayNumber_ = caluculateJulianDayNumber(date);
}

JulianDay::JulianDay(const double julianDayNumber) : julianDayNumber_(julianDayNumber)
{
	;
}

JulianDay::JulianDay(const JulianDay & julianDay) : 
	julianDayNumber_(julianDay.julianDayNumber_)
{
	;
}

JulianDay::~JulianDay()
{
	;
}

wstring JulianDay::asString() const
{
	wstring 	returnValue;
	wchar_t 	tmp[50];
	struct tm 	date = asTm();
	
	size_t len = wcsftime(tmp, 50, L"%m\\%d\\%Y", &date);

	return tmp;
}

double JulianDay::asDouble() const
{
	return julianDayNumber_;
}


struct tm JulianDay::asTm() const 
{
	struct tm date;

	caluculateDate(julianDayNumber_, &date);
	mktime(&date);
	
	return date;
}

void JulianDay::set(double julianDayNumber)
{
	julianDayNumber_ = julianDayNumber;
}

void JulianDay::set(const struct tm * date)
{
	julianDayNumber_ = caluculateJulianDayNumber(date);
}

void JulianDay::set(const wstring & date)
{	
	wstring 	modDate(date);
	struct tm 	dateStruct;

	memset(&dateStruct, 0, sizeof(struct tm));	
	replace(modDate.begin(), modDate.end(), L'\\', L' ');
	wstringstream s(modDate);
	s >> dateStruct.tm_mon >> dateStruct.tm_mday >> dateStruct.tm_year;
	dateStruct.tm_mon -= 1;
	dateStruct.tm_year -= 1900;
	set(&dateStruct);
}

JulianDay & JulianDay::operator= (const JulianDay & rhs)
{
	// Check for assignment to self
	if(this == &rhs) {
		return *this;
	}

	julianDayNumber_ = rhs.julianDayNumber_;
	return *this;
}

JulianDay JulianDay::operator+ (const int days)
{
	JulianDay returnValue(julianDayNumber_ + (double)days);
	return returnValue;
}

void JulianDay::operator+= (int days)
{
	julianDayNumber_ += (double)days;
}

// Prefix
JulianDay & JulianDay::operator++ ()
{
	julianDayNumber_++;
	return *this;
}

// Postfix
JulianDay JulianDay::operator++ (int)
{
	JulianDay returnValue = *this;
	julianDayNumber_++;
	return returnValue;
}

JulianDay JulianDay::operator- (const int days)
{
	JulianDay returnValue(julianDayNumber_ - (double)days);
	return returnValue;
}

void JulianDay::operator-= (const int days)
{
	julianDayNumber_ -= (double)days;
}

// Prefix
JulianDay & JulianDay::operator-- ()
{
	julianDayNumber_--;
	return *this;
}

// Postfix
JulianDay JulianDay::operator-- (int)
{
	JulianDay returnValue = *this;
	julianDayNumber_--;
	return returnValue;
}

bool operator== (const JulianDay & lhs, const JulianDay & rhs) {
	return lhs.julianDayNumber_ == rhs.julianDayNumber_;
}

bool operator< (const JulianDay & lhs, const JulianDay & rhs) {
	return lhs.julianDayNumber_ < rhs.julianDayNumber_;
}

bool operator> (const JulianDay & lhs, const JulianDay & rhs) {
	return lhs.julianDayNumber_ > rhs.julianDayNumber_;
}

bool operator<= (const JulianDay & lhs, const JulianDay & rhs) {
	return lhs.julianDayNumber_ <= rhs.julianDayNumber_;
}

bool operator>= (const JulianDay & lhs, const JulianDay & rhs) {
	return lhs.julianDayNumber_ >= rhs.julianDayNumber_;
}

bool operator!= (const JulianDay & lhs, const JulianDay & rhs) {
	return lhs.julianDayNumber_ != rhs.julianDayNumber_;
}

int operator- (const JulianDay & lhs, const JulianDay & rhs) {
	return lhs.julianDayNumber_ - rhs.julianDayNumber_;
}
