//    Austin is a program for updating projects kept in XML files.
//    Copyright (C) 2000  Joseph Gregorio
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#pragma warning(disable:4786)

#include <iterator>
#include <algorithm>
#include "Project.h"
#include "Utility.h"
#include "dom\DOM_Document.hpp"
#include "DirectedGraph.h"

#ifndef min 
template<class T> inline
const T & min(const T & x, const T & y)	{
	return (x<y ? x : y); 
}
#endif

using namespace std;

Project::Project() {
}

Project::Project(const Project & rhs) {
	name_ = rhs.name_;
	description_ = rhs.description_;
	tasks_ = rhs.tasks_;
}

Project::~Project() {
}

wstring	Project::name() {
	return name_;
}

void	Project::name(const wstring & value) {
	name_ = value;
}

wstring Project::description() {
	return description_;
}

void 	Project::description(const wstring & value) {
	description_ = value;
}

vector<Task> & Project::tasks() {
	return tasks_;
}


struct TaskNameEqualTo {
	wstring name_;
	TaskNameEqualTo(const wstring & name) : name_(name) {}
	bool operator()(Task & task) {
		return (task.name() == name_);
	}
};

Task Project::find(const wstring & name) {
	Task returnValue;
	vector<Task>::iterator loc = find_if(tasks_.begin(), tasks_.end(), TaskNameEqualTo(name));
	if (tasks_.end() != loc) {
		returnValue = *loc;
	}
	return returnValue;
}

Task & Project::findOrCreate(const wstring & name) {
	vector<Task>::iterator loc = find_if(tasks_.begin(), tasks_.end(), TaskNameEqualTo(name));
	if (tasks_.end() == loc) {
		tasks_.push_back(Task());
		loc = tasks_.end()-1;
		loc->name(name);
	} 
	return *loc;
}

DirectedGraph<Task> Project::fillGraph(bool forward) {
	DirectedGraph<Task> taskGraph;
	copy(tasks_.begin(), tasks_.end(), back_inserter(taskGraph));
	for (vector<Task>::iterator i = tasks_.begin(); i != tasks_.end(); ++i) {
		vector<wstring> pred = (*i).predecessors();
		for (vector<wstring>::iterator j = pred.begin(); j != pred.end(); ++j ) {
			Task predTask = find(*j);
			if (forward) {
				taskGraph.insert_edge(predTask, *i);
			} else {
				taskGraph.insert_edge(*i, predTask);
			}
		}
	}

	return taskGraph;	
}

// Add in Start and Finish and add in dependencies if needed.
void Project::rationalize() {
	// Ensure that Start and Finish are present and have 0 times.
	findOrCreate(L"Start");
	findOrCreate(L"Finish");

	DirectedGraph<Task> taskGraph = fillGraph(true);

	Task finish = findOrCreate(L"Finish");
	// Then make sure Finish is the successor of all nodes that do not have a successor.
	for (vector<Task>::iterator i = tasks_.begin(); i != tasks_.end(); ++i) {
		set<Task> successors = taskGraph.adjacency(*i);
		if (0 == successors.size() && (*i) != finish) {
			finish.predecessors().push_back(i->name());
		} 
	}

	
	Task start  = findOrCreate(L"Start");
	// Now rationalize the links, making sure if a node has no predecessors that Start becomes it's pred.
	for (i = tasks_.begin(); i != tasks_.end(); ++i) {
		if (0 == i->predecessors().size()) {
			if (start != *i) {
				i->predecessors().push_back(L"Start");
			}
		}
	}	
}

struct CompareTasksByFinish {
	bool operator()(Task & lhs, Task & rhs) {
		return lhs.finish() < rhs.finish();
	}
};

struct CompareTasksByLateStart {
	bool operator()(Task & lhs, Task & rhs) {
		return lhs.lateStart() < rhs.lateStart();
	}
};

void updateForPercentComplete(Task task, bool optimistic) {
	JulianDay today;
	if (task.percentComplete() < 1.0f) {	
		task.finish(task.start() + task.duration());
		if (task.work() > 0) {
			double calcPercentComplete = (double)(today-task.start())/(double)task.duration();
			if (calcPercentComplete > task.percentComplete()) {
				if (optimistic) {
					task.percentComplete(min<double>(1.0f, calcPercentComplete));
				} else {
					task.finish(today + (1-task.percentComplete())*task.duration());
				}
			}
		}
	}	
}

vector<Task> Project::tasksPartiallyOrdered() {
	DirectedGraph<Task> taskGraphReverse = fillGraph(false);
	vector<Task> tasksDFS;
	vector<Task> loop = dfs(taskGraphReverse, back_inserter(tasksDFS));

	return tasksDFS;
}

void Project::calculateStartAndFinishTimes(bool optimistic) {
	DirectedGraph<Task> taskGraph = fillGraph(true);
	DirectedGraph<Task> taskGraphReverse = fillGraph(false);
	

	// Fill in ES, EF
	vector<Task> tasksDFS;
	vector<Task> loop = dfs(taskGraphReverse, back_inserter(tasksDFS));
	if (0 == loop.size() && 0 != tasksDFS.size()) {
		Task start = tasksDFS[0];
		updateForPercentComplete(start, optimistic);
		for (vector<Task>::iterator i = tasksDFS.begin()+1; i != tasksDFS.end(); ++i) {
			set<Task> predecessors = taskGraph.predecessors(*i);
			Task maxPred = *max_element(predecessors.begin(), predecessors.end(), CompareTasksByFinish());
			if (0 == i->percentComplete()) {
				i->start(maxPred.finish());
			}
			updateForPercentComplete(*i, optimistic);
		}
	} else {
		cout << "Cycle in the dependencies, or there were no nodes to process." << endl;
	}

	// Fill in LS, LF
	vector<Task> tasksReverseDFS;
	loop = dfs(taskGraph, back_inserter(tasksReverseDFS));
	if (0 == loop.size() && 0 != tasksReverseDFS.size()) {
		Task finish = tasksReverseDFS[0];
		finish.lateStart(finish.finish());
		finish.lateFinish(finish.finish());		
		for (vector<Task>::iterator i = tasksReverseDFS.begin()+1; i != tasksReverseDFS.end(); ++i) {			
			set<Task> predecessors = taskGraphReverse.predecessors(*i);
			Task minPred = *min_element(predecessors.begin(), predecessors.end(), CompareTasksByLateStart());
			i->lateFinish(minPred.lateStart());
			i->lateStart(i->lateFinish() - i->duration());
		}
	} else {
		cout << "Cycle in the dependencies, or there were no nodes to process." << endl;
	}
}

struct DOMTaskToTask {
	Task operator()(DOM_Element taskElement) {
		Task returnValue;
		returnValue.readConfig(taskElement);
		return returnValue;
	}
};

struct TaskToName {
	wstring operator()(Task & task) {
		return task.name();
	}
};

void Project::readConfig(DOM_Element project) {
	tasks_.clear();
	vector<DOM_Element> tasks = collectElements(L"task", project);
	transform(tasks.begin(), tasks.end(), back_inserter(tasks_), DOMTaskToTask());

	description_ = getSingleElementByName(project, L"description").getFirstChild().getNodeValue().rawBuffer();
	name_ = getSingleElementByName(project, L"name").getFirstChild().getNodeValue().rawBuffer();
}

// Note that this just does the insert, not the write.
struct InsertTask {
	DOM_Element project_;
	InsertTask(DOM_Element project) : project_(project) {}
	void operator()(wstring & taskName) {
		DOM_Element taskNode = project_.getOwnerDocument().createElement(L"task");
		taskNode.setAttribute(L"id", taskName.c_str());
		project_.appendChild(taskNode);
	}
};

struct RemoveTask {
	DOM_Element project_;
	RemoveTask(DOM_Element project) : project_(project) {}
	void operator()(wstring & taskName) {
		vector<DOM_Element> tasks = collectElements(L"task", project_);
		vector<DOM_Element>::iterator taskNodeIt = find_if(tasks.begin(), tasks.end(), HasID(taskName.c_str()));
		if (tasks.end() != taskNodeIt) {
			project_.removeChild(*taskNodeIt);
		}
	}
};


struct WriteTaskToDOM {
	vector<DOM_Element> tasks_;
	WriteTaskToDOM(DOM_Element project) {
		tasks_ = collectElements(L"task", project);
	}
	void operator()(Task & task) {
		vector<DOM_Element>::iterator taskNodeIt = find_if(tasks_.begin(), tasks_.end(), HasID(task.name().c_str()));
		if (tasks_.end() != taskNodeIt) {
			task.writeConfig(*taskNodeIt);
		}
	}
};


void Project::writeConfig(DOM_Element project) {
	getSingleElementByName(project, L"name").getFirstChild().setNodeValue(name_.c_str());
	getSingleElementByName(project, L"description").getFirstChild().setNodeValue(description_.c_str());


	// Update the list of tasks. Add in the ones added and remove the ones we have dropped.
	vector<wstring> taskNames;
	transform(tasks_.begin(), tasks_.end(), back_inserter(taskNames), TaskToName());
	vector<DOM_Element> orig_tasks = collectElements(L"task", project);
	vector<wstring> orig_task_names;
	transform(orig_tasks.begin(), orig_tasks.end(), back_inserter(orig_task_names), ElementToName());
	// We need to sort to do set_difference().
	sort(orig_task_names.begin(), orig_task_names.end());
	sort(taskNames.begin(), taskNames.end());
	vector<wstring> toBeRemoved;
	vector<wstring> toBeInserted;
	// Determine which tasks have been added and removed.
	set_difference(orig_task_names.begin(), orig_task_names.end(), taskNames.begin(), taskNames.end(), back_inserter(toBeRemoved));
	set_difference(taskNames.begin(), taskNames.end(), orig_task_names.begin(), orig_task_names.end(), back_inserter(toBeInserted));
	// Now change the DOM to reflect the changes.
	for_each(toBeInserted.begin(), toBeInserted.end(), InsertTask(project));
	for_each(toBeRemoved.begin(), toBeRemoved.end(), RemoveTask(project));	

	// Now write out the tasks.
	for_each(tasks_.begin(), tasks_.end(), WriteTaskToDOM(project));
}
	











